require("defines")
local serpent = require("serpent")

function string.starts(String,Start)
  return string.sub(String,1,string.len(Start))==Start
end
function Set(l)
  local s={}
  for i,k in ipairs(l) do
    s[k]=i
  end
  return s
end

function getTileBBox(p)
  return{
    {math.floor(p.x),math.floor(p.y)},
    {math.ceil(p.x),math.ceil(p.y)}
  }
end
function getBBox(p,r)
  return{{p.x-r,p.y-r},{p.x+r,p.y+r}}
end
function hasinventory(e,i)
  if i == nil then
    for _,k in pairs(defines.inventory) do
      if pcall(function()e.get_inventory(k)end) then
        return e.get_inventory(k)
      end
    end
  elseif pcall(function()e.get_inventory(i)end) then
    return e.get_inventory(i)
  end
  return false
end
function hasfluidbox(e)
  if pcall(function() return e.fluidbox end) then
    return #e.fluidbox
  end
  return false
end


ground_entity = Set{
  "item-entity",
  "simple-entity",
  "transport-belt",
  "transport-belt-to-ground",
  "decorative",
  "smoke",
  "construction-robot",
  "logistic-robot",
  "player"
}
chest_output = {
  [0]={x1=-0.25,y1=-0.75,x2= 0.25,y2=-0.75},
  [2]={x1= 0.75,y1=-0.25,x2= 0.75,y2= 0.25},
  [4]={x1=-0.25,y1= 0.75,x2= 0.25,y2= 0.75},
  [6]={x1=-0.75,y1=-0.25,x2=-0.75,y2= 0.25}
}
chest_input = {
  [0]={x1=-0.25,y1=-0.51,x2= 0.25,y2=-0.51},
  [2]={x1= 0.51,y1=-0.25,x2= 0.51,y2= 0.25},
  [4]={x1=-0.25,y1= 0.51,x2= 0.25,y2= 0.51},
  [6]={x1=-0.51,y1=-0.25,x2=-0.51,y2= 0.25}
}

function on_init()
  if not global.item_source then global.item_source = {} end
  if not global.item_sink then global.item_sink = {} end
  if not global.energy_source then global.energy_source = {} end
  if not global.energy_sink then global.energy_sink = {} end
end
function on_built_entity(event)
  if event.created_entity.name == "item-source" then
    table.insert(global.item_source, event.created_entity)
  elseif event.created_entity.name == "item-sink" then
    table.insert(global.item_sink, event.created_entity)
  elseif event.created_entity.name == "energy-source" then
    table.insert(global.energy_source, event.created_entity)
  elseif event.created_entity.name == "energy-sink" then
    table.insert(global.energy_sink, event.created_entity)
  end
end
function on_tick(event)
  for index, item_sink in ipairs(global.item_sink) do
    if item_sink.valid then
      item_sink.energy = 10
      local pos = item_sink.position
      local dir = chest_input[item_sink.direction]
      input{x = pos.x + dir.x1, y = pos.y + dir.y1}
      inputfluid{x = pos.x + dir.x1, y = pos.y + dir.y1}
      input{x = pos.x + dir.x2, y = pos.y + dir.y2}
      inputfluid{x = pos.x + dir.x2, y = pos.y + dir.y2}
    else
      table.remove(global.item_sink, index)
    end
  end
  for index, item_source in ipairs(global.item_source) do
    if item_source.valid then
      item_source.energy = 10
      local slot1 = {name = item_source.get_filter(1), count = 1}
      local slot2 = {name = item_source.get_filter(2), count = 1}
      local pos = item_source.position
      local dir = chest_output[item_source.direction]
      if slot1.name ~= nil then
        if string.starts(slot1.name,"creative-fluid-") then
          outputfluid({type=string.sub(slot1.name,16),amount=10}, {x = pos.x + dir.x1, y = pos.y + dir.y1})
        else
          output(slot1, {x = pos.x + dir.x1, y = pos.y + dir.y1})
        end
      end
      if slot2.name ~= nil then
        if string.starts(slot2.name,"creative-fluid-") then
          outputfluid({type=string.sub(slot2.name,16),amount=10}, {x = pos.x + dir.x1, y = pos.y + dir.y1})
        else
          output(slot2, {x = pos.x + dir.x2, y = pos.y + dir.y2})
        end
      end
    else
      table.remove(global.item_source, index)
    end
  end
  if game.tick % 2 == 0 then
    for index, energy_sink in ipairs(global.energy_sink) do
      if energy_sink.valid then
        energy_sink.energy = 0
      else
        table.remove(global.energy_sink, index)
      end
    end
    for index, energy_source in ipairs(global.energy_source) do
      if energy_source.valid then
        energy_source.energy = 1000000000000000
      else
        table.remove(global.energy_source, index)
      end
    end
  end
end
function output(item,position)
  local storage_entity = nil
  local on_ground = true
  for _, entity in ipairs(game.surfaces['nauvis'].find_entities(getTileBBox(position))) do
    if not ground_entity[entity.type] then
      on_ground = false
      if entity.can_insert(item) then
        storage_entity = entity
        break
      end
    end
  end
  if storage_entity then
    local buf = storage_entity.get_item_count(item.name)
    storage_entity.insert(item)
    local buf2 = storage_entity.get_item_count(item.name)
    item.count = buf2-buf
    return item
  end
  if on_ground and not next(game.surfaces['nauvis'].find_entities_filtered{area=getBBox(position, 0.14), name="item-on-ground"}) then
    game.surfaces['nauvis'].create_entity{name = "item-on-ground", position = position, stack = item}
  end
end
function input(position)
  local storage_entity = nil
  local on_ground = true
  for _, entity in ipairs(game.surfaces['nauvis'].find_entities(getTileBBox(position))) do
    if not ground_entity[entity.type] and entity.get_item_count() > 0 then
      storage_entity = entity
      break
    end
  end
  if storage_entity then
    for _, i in pairs(defines.inventory) do
      local inventory = hasinventory(storage_entity,i)
      if inventory and not inventory.is_empty() then
        for i = 1,#inventory do
          if inventory[i] ~= nil then
            item = {name=inventory[i].name,count=1}
            inventory.remove(item)
            return item
          end
        end
      end
    end
  end
  items = game.surfaces['nauvis'].find_entities_filtered{area=getBBox(position,0.14),name="item-on-ground"}
  for _,item in ipairs(items) do
    item.destroy()
  end
end
function outputfluid(item,position)
  local storage_entity = nil
  for _, entity in ipairs(game.surfaces['nauvis'].find_entities(getTileBBox(position))) do
    if not ground_entity[entity.type] and hasfluidbox(entity) then
      storage_entity = entity
      break
    end
  end
  if storage_entity then
    local fluidbox = storage_entity.fluidbox
    if fluidbox then
      for i = 1,#fluidbox do
        fluid = fluidbox[i]
        if fluid == nil then
          fluidbox[i] = {type = item.type, amount = item.amount}
        elseif fluid.type == item.type then
          fluid.amount = fluid.amount + item.amount
          fluidbox[i] = fluid
        end
      end
    end
  end
end
function inputfluid(position)
  local storage_entity = nil
  local on_ground = true
  for _, entity in ipairs(game.surfaces['nauvis'].find_entities(getTileBBox(position))) do
    if not ground_entity[entity.type] and hasfluidbox(entity) then
      storage_entity = entity
      break
    end
  end
  if storage_entity then
    local fluidbox = storage_entity.fluidbox
    if fluidbox then
      for i = 1,#fluidbox do
        amount = 1
        fluid = fluidbox[i]
        if fluid then
          if amount >= fluid.amount then
            amount = fluid.amount
            fluidbox[i] = nil
          else
            fluid.amount = fluid.amount - amount
            fluidbox[i] = fluid
          end
          return {type = fluid, amount = amount, temperature = fluid.temperature}
        end
      end
    end
  end
end

on_load = on_init
on_robot_built_entity = on_built_entity

game.on_init(on_init)
game.on_load(on_load)
game.on_event(defines.events,on_event)
for k,v in pairs(defines.events) do
  if type(_G[k]) == "function" then
    game.on_event(v,_G[k])
  end
end